/*	Dean's Personal - May 1989	*/
#ident "@(#)bu.c	1.0 bu.c 10/13/88 18:27:34"

/* bu [-mq] filename ...
 * creates filename.NN as a copy of filename.
 * filename.NN will have the same mod date and rx permissions
 * as filename, but will not have write permission.
 */

/*
 * For each filename
 *	verify that file exists, is a normal file, and is readable
 *	Verify that directory is writeable
 *	truncate last component of filename to 11 characters.
 *	search directory for filename.NN filenames, find largest
 *		NN, and increment to form new filename.
 *	Copy filename to new filename, reset permissions, mod date
 *		group and ownership
 *	if !quiet print out new filename
 *
 * NOTES:
 *	Error messages go to stderr
 *	new filename goes to stdout
 */

#include <string.h>
#include <sys/types.h>
#include <dirent.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <memory.h>
#include <stdio.h>
#include <signal.h>
#include <varargs.h>
#include <sys/stat.h>
#include <time.h>
#include <malloc.h>
#include <utime.h>

#define TRUE 1
#define FALSE 0

#define COPYSIZE 1024
#define DIRPATHMAX 511
#define FNAMEMAX 32

char	dirname[DIRPATHMAX+1], fname[FNAMEMAX+1];
int prdirname;
int euid, egid;	/* Effective UID and GID */
char copybuf[COPYSIZE];

extern	errno;
extern	void exit();
extern	long ulimit();
char *myname;
int fdorig, fdbu;

int Quiet, Move;

main(argc, argv)
char **argv;
{
	extern char	*optarg;
	extern int	optind;
	int ans;
	int exitcode;

	exitcode = 0;			/* Default exit code is AOK */
	myname = *argv;
	fname[FNAMEMAX] = '\0';
	euid = geteuid();
	egid = getegid();

/*	printf("euid = %d, egid = %d\n", euid, egid);	*/

	while( (ans = getopt( argc, argv, "qm")) != EOF ) {

		switch( ans ) {

		case 'q':		/* Don't print message */
			Quiet = TRUE;
			break;

		case 'm':
			Move = TRUE;	/* Move, don't copy */
			break;

		default:
			usage();
			exit(1);
		}
	}

	argc -= optind;
	argv += optind;

	if (argv[0] == NULL) {
		usage();
		exit(1);
	}

	while(argv[0] != NULL) {
			/* Do the backup, if ANY attempt fails, return a non
			 * zero exit code
			 */
		if (backup(argv[0]) != TRUE) {
			exitcode = 1;
		}
		argv++;
	}

	exit(exitcode);
}

usage() {
	printf("Usage: %s [-mq] filename ...\n", myname);
}

/*
 *	verify that file exists, is a normal file, and is readable
 *	Verify that directory is writeable
 *	truncate last component of filename to 11 characters.
 *	search directory for filename.NN filenames, find largest
 *		NN, and increment to form new filename.
 *	Copy filename to new filename, reset permissions, mod date
 *		group and ownership
 *	if !quiet print out new filename
 */

backup(name)
char *name;
{
	int rval, bytecount, bunum;
	char *slash;
	int l, retcode;
	struct stat sbuf;
	char *newname;
	struct utimbuf timebuf;

	retcode = FALSE;

	if ((fdorig = open(name, O_RDONLY) ) < 0) {
		errmsg(name,NULL);
		return(FALSE);
	}

		/* Get the stat information needed:
		 *		mode
		 *		uid, gid
		 *		mod date
		 *		permisions
		 */

	if ((rval = fstat(fdorig, &sbuf) ) < 0) {
		errmsg(name,NULL);
		goto errout;
	}

/*	printf("file uid is %d, gid is %d, mode is 0%o\n", sbuf.st_uid,
/*			sbuf.st_gid, sbuf.st_mode);
/*	printf("\tMod time is %s", ctime(&sbuf.st_mtime));
 */
	if ((sbuf.st_mode & S_IFMT) != S_IFREG) {
		errmsg(name, "Not an ordinary file");
		goto errout;
	}

		/* Generate the name use for the directory */
	if ((slash = strrchr(name, '/')) == NULL) {
			/* Gee, there is no directory name */
		prdirname = FALSE;
		dirname[0] = '.';
		dirname[1] = '/';
		dirname[2] = '\0';
		strncpy(fname, name, FNAMEMAX);
	}else{
		prdirname = TRUE;
		if ((slash - name) > DIRPATHMAX) {
			errmsg(name, "Directory name too long");
			goto errout;
		}

			/* dirname will INCLUDE the /, just so "bu /unix" will work */

		strncpy(dirname, name, slash - name + 1);
		dirname[slash - name + 1] = '\0';
		slash++;
		strncpy(fname, slash, FNAMEMAX);
	}
		/* Truncate the name, if it is too long */
	fname[FNAMEMAX - 3] = '\0';
	if ((bunum = findbunum(dirname, fname)) == -1) {
		goto errout;
	}
	l = strlen(fname);
	fname[l++] = '.';
	fname[l++] = ((bunum / 10) % 10) + '0';
	fname[l++] = (bunum % 10) + '0';
	fname[l] = '\0';

		/* Now copy the file */
		/* 1) Generate the name */

	newname = malloc(strlen(dirname) + strlen(fname) + 1);
	strcpy(newname, dirname);
	strcat(newname, fname);

	if (Move) {	/* Just move the file, don't copy */
		if (link(name, newname) != 0) {
			errmsg(newname, "Cannot link");
			free(newname);
			goto errout;
		}
		if (unlink(name) != 0) {
			errmsg(name, "Cannot unlink");
			free(newname);
			goto errout;
		}

		if (chmod(newname, sbuf.st_mode &
			  ~(S_IWUSR | S_IWGRP | S_IWOTH)) != 0) {
			errmsg(newname, NULL);
			free(newname);
			goto errout;
		}
		goto okout;
	}

		/* 2) create the file */

	if ((fdbu = creat(newname, 0600)) == -1) {
		errmsg(newname, NULL);
		free(newname);
		goto errout;
	}

		/* 3) copy the file */
	while ((bytecount = read(fdorig, copybuf, COPYSIZE)) > 0) {
		if ((rval = write(fdbu, copybuf, bytecount)) != bytecount) {
			errmsg(newname, NULL);
			fprintf(stderr, "\tCannot write %s\n", newname);
			close(fdbu);
			if (unlink(newname) != 0) {
				fprintf(stderr, "\tCannot remove %s\n", newname);
			}
			free(newname);
			goto errout;
		}
	}

	close(fdbu);

	if (bytecount < 0) {
		errmsg(name, NULL);
		if (unlink(newname) != 0) {
			fprintf(stderr, "\tCannot remove %s\n", newname);
		}
		free(newname);
		goto errout;
	}

	timebuf.actime = sbuf.st_atime;
	timebuf.modtime = sbuf.st_mtime;

	if ( utime(newname, &timebuf) != 0 ) {
		errmsg(newname, NULL);
		if (unlink(newname) != 0) {
			fprintf(stderr, "\tCannot remove %s\n", newname);
		}
		free(newname);
		goto errout;
	}

	if (chmod(newname, sbuf.st_mode & ~(S_IWUSR | S_IWGRP | S_IWOTH)) != 0) {
		errmsg(newname, NULL);
		if (unlink(newname) != 0) {
			fprintf(stderr, "\tCannot remove %s\n", newname);
		}
		free(newname);
		goto errout;
	}

	if ( (sbuf.st_uid != euid) || (sbuf.st_gid != egid) ) {
		if (chown(newname, sbuf.st_uid, sbuf.st_gid) != 0) {
			errmsg(newname, NULL);
			if (unlink(newname) != 0) {
				fprintf(stderr, "\tCannot remove %s\n", newname);
			}
			free(newname);
			goto errout;
		}
	}

	free(newname);

okout:
	if (! Quiet) {
		if (prdirname) {
			printf("%s", dirname);
		}
		printf("%s\n", fname);
	}

	retcode = TRUE;

errout:
	close(fdorig);
	return(retcode);
}

errmsg(name,msg)
char *name, *msg;
{
	fprintf(stderr, "%s: %s: ", myname, name);
	if (msg != NULL) {
		fprintf(stderr, "%s\n", msg);
	}else{
		perror("");
	}
}

/*
 * In directory dir, find files of the form <file>.NN
 * Where N is [0-9]
 * return the largest value of NN + 1
 *	Open the directory
 *	for each entry
 *		if entry == file (up to last character of file)
 *			if next four chars of entry are .[0-9][0-9]\0
 *				figure out what NN is
 *				if new NN is greater than any previous NN note it
 *	return largest previous value plus 1
 *	return -1 on any error and print out a message
 */

findbunum(dir, file)
char *dir;
char *file;
{
	int bunum, num;
	DIR *dirp;
	struct dirent *pde;
	register int len;

	bunum = -1;
	len = strlen(file);

	if ((dirp = opendir(dir)) == NULL) {
		errmsg(dir, NULL);
		return(-1);
	}
		/* See if a directory entry matches */
	 while ((pde = readdir(dirp)) != NULL) {
		if (strncmp(pde->d_name, file, len) == 0) {
				/* They match enough! */
			if ((pde->d_name[len] == '.') &&
					(isdigit((int)pde->d_name[len+1])) &&
					(isdigit((int)pde->d_name[len+2])) &&
					(pde->d_name[len + 3] == '\0') ) {
				num = (((int)pde->d_name[len+1] - '0') * 10) +
						((int)pde->d_name[len+2] - '0');
				if (num >= bunum) bunum = num;
			}
		}
	}

	bunum++;

	if (bunum >= 100) {
		fprintf(stderr, "%s: ", myname);
		if (prdirname) {
			fprintf(stderr, "%s", dir);
		}
		fprintf(stderr, "%s.99 already exists! Cannot back up file.\n",
			file);
		bunum = -1;
	}

	closedir(dirp);
	return(bunum);
}
